from enum import Enum
from ..names import Names

class BMPType(Enum):
    """The code will be the one used in the database"""
    BMP_TYPE_POINTSOURCE = 1
    BMP_TYPE_FLOWDIVERSION_STREAM = 2
    BMP_TYPE_RESERVOIR = 3
    BMP_TYPE_RIPARIANWETLAND = 4
    BMP_TYPE_RIPARIANBUFFER = 5
    BMP_TYPE_GRASSWATERWAY = 6
    BMP_TYPE_FILTERSTRIP = 7
    BMP_TYPE_POND = 8
    BMP_TYPE_WETLAND = 9
    BMP_TYPE_TERRACE = 10
    BMP_TYPE_FLOWDIVERSION_OVERLAND = 11
    BMP_TYPE_CROP = 12
    BMP_TYPE_RESIDUAL = 13
    BMP_TYPE_TILLAGE = 14
    BMP_TYPE_FERTILIZER = 15
    BMP_TYPE_GRAZING = 16
    BMP_TYPE_PESTICIDE = 17
    BMP_TYPE_IRRIGATION = 18
    BMP_TYPE_TILEDRAIN = 19
    BMP_TYPE_URBAN = 20
    BMP_TYPE_MI48H = 21
    BMP_TYPE_MSETBACK = 22
    BMP_TYPE_NO_ONSNOW = 23
    BMP_TYPE_NO_FALL = 24
    BMP_TYPE_NITROGEN_LIMIT = 25
    BMP_TYPE_PHOSPHORUS_LIMIT = 26
    BMP_TYPE_MANURE_STORAGE = 27
    BMP_TYPE_MANURE_CATCHBASIN = 28
    BMP_TYPE_MANURE_FEEDLOT = 29
    BMP_TYPE_CROP_MAR = 30
    BMP_TYPE_FERTILIZER_MAR = 31
    BMP_TYPE_TILLAGE_MAR = 32
    BMP_TYPE_WINTERING_SITE = 33
    BMP_TYPE_CROP_PS = 34
    BMP_TYPE_FERTILIZER_PS = 35
    BMP_TYPE_TILLAGE_PS = 36
    BMP_TYPE_GRAZING_PS = 37
    BMP_TYPE_DUGOUT = 38
    BMP_TYPE_OFFSITEWATERING = 39
    BMP_TYPE_ACCESSMGT = 40
    BMP_TYPE_WASCOB = 41
    BMP_TYPE_WATERUSE = 42

ReachBMPColumnNames = {
    BMPType.BMP_TYPE_POINTSOURCE: "PointSource",
    BMPType.BMP_TYPE_FLOWDIVERSION_STREAM: "FlowDiversion",
    BMPType.BMP_TYPE_RESERVOIR: "Reservoir",
    BMPType.BMP_TYPE_WETLAND:"Wetland",
    BMPType.BMP_TYPE_MANURE_CATCHBASIN:"CatchBasin",
    BMPType.BMP_TYPE_GRASSWATERWAY:"GrassWaterway",
    BMPType.BMP_TYPE_ACCESSMGT:"AccessManagement",
    BMPType.BMP_TYPE_WATERUSE:"WaterUse"
}

ArealNonStructureBMPs = [
    BMPType.BMP_TYPE_CROP,
    BMPType.BMP_TYPE_FERTILIZER,
    BMPType.BMP_TYPE_TILLAGE,
    BMPType.BMP_TYPE_GRAZING,
    BMPType.BMP_TYPE_MANURE_STORAGE,
    BMPType.BMP_TYPE_MANURE_FEEDLOT
]

ArealStructureBMPs = [
    BMPType.BMP_TYPE_DUGOUT,
    BMPType.BMP_TYPE_RIPARIANBUFFER,
    BMPType.BMP_TYPE_FILTERSTRIP,
    BMPType.BMP_TYPE_TILEDRAIN,
    BMPType.BMP_TYPE_WASCOB
]

#bmp distribution
BMPDistributions = {
    #reach bmps use the same distriubtion table reach_bmp
    BMPType.BMP_TYPE_POINTSOURCE: Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_FLOWDIVERSION_STREAM: Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_RESERVOIR: Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_WETLAND:Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_MANURE_CATCHBASIN:Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_GRASSWATERWAY:Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_DUGOUT:Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_ACCESSMGT:Names.bmp_table_name_reach_bmp,
    BMPType.BMP_TYPE_WATERUSE:Names.bmp_table_name_reach_bmp,

    #non-structure:[distribution_raster_name_in_hdf5]/[management_table_name]
    #use field for crop, fertilizer, tillage and grazing
    BMPType.BMP_TYPE_CROP:f"{Names.remove_extension_from_raster_file(Names.fieldRasName)}/{Names.bmp_table_name_crop_management}",
    BMPType.BMP_TYPE_FERTILIZER:f"{Names.remove_extension_from_raster_file(Names.fieldRasName)}/{Names.bmp_table_name_fertilizer_management}",
    BMPType.BMP_TYPE_TILLAGE:f"{Names.remove_extension_from_raster_file(Names.fieldRasName)}/{Names.bmp_table_name_tillage_management}",
    BMPType.BMP_TYPE_GRAZING:f"{Names.remove_extension_from_raster_file(Names.fieldRasName)}/{Names.bmp_table_name_grazing_management}",
    BMPType.BMP_TYPE_MANURE_STORAGE: f"{Names.remove_extension_from_raster_file(Names.manureStorageRasName)}/{Names.bmp_table_name_manure_storage_management}",
    BMPType.BMP_TYPE_MANURE_FEEDLOT: f"{Names.remove_extension_from_raster_file(Names.feedlotBoundaryProcessedRasName)}/{Names.bmp_table_name_manure_feed_lot_management}",     

    #structure
    #riparian buffer and filter strip has two distribution raster: part and drainage
    #other structure bmps have only one distribution raster
    BMPType.BMP_TYPE_RIPARIANBUFFER: f"{Names.remove_extension_from_raster_file(Names.riparianBufferPartRasterName)}/{Names.remove_extension_from_raster_file(Names.riparianBufferDrainageRasterName)}",  
    BMPType.BMP_TYPE_FILTERSTRIP: f"{Names.remove_extension_from_raster_file(Names.filterStripPartRasterName)}/{Names.remove_extension_from_raster_file(Names.filterStripDrainageRasterName)}", 
    BMPType.BMP_TYPE_DUGOUT:Names.remove_extension_from_raster_file(Names.dugoutRasName),
    BMPType.BMP_TYPE_WASCOB:Names.remove_extension_from_raster_file(Names.wascobRasName),
    BMPType.BMP_TYPE_TILEDRAIN:Names.remove_extension_from_raster_file(Names.tileDrainRasName),


    #manure application adjustment:[distribution_raster_name_in_hdf5]/[management_table_name]
    #they may share the same distribution raster, maybe field?
    BMPType.BMP_TYPE_MI48H :  f"{Names.remove_extension_from_raster_file(Names.manure48hRasName)}/{Names.bmp_table_name_manure_incorporation_within_48h_management}",
    BMPType.BMP_TYPE_MSETBACK :  f"{Names.remove_extension_from_raster_file(Names.manureSetbackRasName)}/{Names.bmp_table_name_manure_application_setback_management}",
    BMPType.BMP_TYPE_NO_ONSNOW :  f"{Names.remove_extension_from_raster_file(Names.manureNoOnSnowRasName)}/{Names.bmp_table_name_manure_no_application_on_snow_management}",
    BMPType.BMP_TYPE_NO_FALL :  f"{Names.remove_extension_from_raster_file(Names.manureSpringRasName)}/{Names.bmp_table_name_manure_spring_application_rather_than_fall_application_management}",
    BMPType.BMP_TYPE_NITROGEN_LIMIT :  f"{Names.remove_extension_from_raster_file(Names.manureNLimitRasname)}/{Names.bmp_table_name_manure_application_based_on_soil_nitrogen_limit_management}",
    BMPType.BMP_TYPE_PHOSPHORUS_LIMIT :  f"{Names.remove_extension_from_raster_file(Names.manurePLimitRasname)}/{Names.bmp_table_name_manure_application_based_on_soil_phosphorous_limit_management}"
}

#bmp parameters
BMPParameters = {
    #reach bmps
    BMPType.BMP_TYPE_POINTSOURCE: Names.bmp_table_name_point_source,
    BMPType.BMP_TYPE_FLOWDIVERSION_STREAM: Names.bmp_table_name_flow_diversion,
    BMPType.BMP_TYPE_RESERVOIR: Names.bmp_table_name_reservoir,
    BMPType.BMP_TYPE_WETLAND:Names.bmp_table_name_wetland,
    BMPType.BMP_TYPE_MANURE_CATCHBASIN:Names.bmp_table_name_manure_catch_basin,
    BMPType.BMP_TYPE_GRASSWATERWAY:Names.bmp_table_name_grass_waterway,
    BMPType.BMP_TYPE_DUGOUT:Names.bmp_table_name_dugout,
    BMPType.BMP_TYPE_ACCESSMGT:Names.bmp_table_name_managed_access_including_fencing,
    BMPType.BMP_TYPE_WATERUSE:Names.bmp_table_name_water_use,

    #non-structure
    #all non-structure bmps has only one parameter table except grazing which has three parameter table:[livestock_parameter]/[fertilizer_parameter]/[offsite_watering_parameter]
    BMPType.BMP_TYPE_CROP:Names.bmp_table_name_crop_parameter,
    BMPType.BMP_TYPE_FERTILIZER:Names.bmp_table_name_fertilizer_parameter,
    BMPType.BMP_TYPE_TILLAGE:Names.bmp_table_name_tillage_parameter,    
    BMPType.BMP_TYPE_MANURE_STORAGE: Names.bmp_table_name_manure_storage_parameter,
    BMPType.BMP_TYPE_MANURE_FEEDLOT: Names.bmp_table_name_manure_feed_lot_parameter,
    BMPType.BMP_TYPE_GRAZING: f"{Names.bmp_table_name_livestock_parameter}/{Names.bmp_table_name_fertilizer_parameter}/{Names.bmp_table_name_offsite_watering}",

    #structure
    BMPType.BMP_TYPE_RIPARIANBUFFER: f"{Names.bmp_table_name_riparian_buffer}/{Names.bmp_table_name_crop_remove_parameter}",
    BMPType.BMP_TYPE_FILTERSTRIP: f"{Names.bmp_table_name_filter_strip}/{Names.bmp_table_name_crop_remove_parameter}",
    BMPType.BMP_TYPE_DUGOUT: Names.bmp_table_name_dugout,
    BMPType.BMP_TYPE_WASCOB:Names.bmp_table_name_wascob,
    BMPType.BMP_TYPE_TILEDRAIN:Names.bmp_table_name_tile_drain,

    #manure application adjustment
    #there is no parameter tables for maure application adjustment bmps
    BMPType.BMP_TYPE_MI48H: "",
    BMPType.BMP_TYPE_MSETBACK: "",
    BMPType.BMP_TYPE_NO_ONSNOW: "",
    BMPType.BMP_TYPE_NO_FALL : "",
    BMPType.BMP_TYPE_NITROGEN_LIMIT: "",
    BMPType.BMP_TYPE_PHOSPHORUS_LIMIT: ""

}

DefaultScenarioId = 2


from.bmp_14_tillage_management import TillageManagement

class Pasture_tillage_management(TillageManagement):
    """Distribution Table for BMP: Pasture tillage management (36)"""
    def __init__(self, TillMon, TillDay, TillCode):
        super().__init__(TillMon, TillDay, TillCode)



class StructureAttribute:
    def __init__(self, id:int, area_ha:float, subbasin:int, contribution_area_ha:float):
        self.id = id
        self.area_ha = area_ha
        self.subbasin = subbasin
        self.contribution_area = contribution_area_ha


from .bmp_reach import ReachBMP
from ..database.bmp.bmp_03_reservoir import Reservoir
from ..vector_extension import VectorExtension
from ..names import Names
from whitebox_workflows import FieldData, Vector, Raster
import os
import pandas as pd

class ReachBMPReservoir(ReachBMP):
    FLOW_ROUTING_NAME_RATING_CURVE = "RAT_RES"
    FLOW_ROUTING_NAME_DAILY_OUTFLOW = "MDO_RES"
    FLOW_ROUTING_NAME_MONTHLY_OUTFLOW = "MMO_RES"
    FLOW_ROUTING_NAME_ANUNAL_RELEASE_RATE = "AAR_RES"
    FLOW_ROUTING_NAME_TARGET_RELEASE_RATE = "TRR_RES"

    SEDIMENT_ROUTING_NAME_MASS_BALANCE = "SMB_RES"
    NUTRIENT_ROUTING_NAME_MASS_BALANCE = "RES_NUTR_BAL"

    FLOW_ROUTING_METHODS = [
            FLOW_ROUTING_NAME_RATING_CURVE,
            FLOW_ROUTING_NAME_DAILY_OUTFLOW,
            FLOW_ROUTING_NAME_MONTHLY_OUTFLOW,
            FLOW_ROUTING_NAME_ANUNAL_RELEASE_RATE,
            FLOW_ROUTING_NAME_TARGET_RELEASE_RATE
    ]

    FLOW_ROUTING_METHODS_REQUIRING_EXTERNAL_DATA = [
            FLOW_ROUTING_NAME_RATING_CURVE,
            FLOW_ROUTING_NAME_DAILY_OUTFLOW,
            FLOW_ROUTING_NAME_MONTHLY_OUTFLOW
    ]

    RATING_CURVE_COLUMNS = [
        "storage","area","discharge"
    ]

    field_name_reservoir_operation = "op"
    field_name_reservoir_emergency_surface_area = "sa_em"
    field_name_reservoir_emergency_volume = "v_em"
    field_name_reservoir_principal_surface_area = "sa_pr"
    field_name_reservoir_principal_volume = "v_pr"
    field_name_reservoir_flow_data = "file"

    fields_reservoir = [
        field_name_reservoir_operation,
        field_name_reservoir_emergency_surface_area,
        field_name_reservoir_emergency_volume,
        field_name_reservoir_principal_surface_area,
        field_name_reservoir_principal_volume,
        field_name_reservoir_flow_data
    ]

    def __init__(self, reservoir_vector:Vector, 
                 subbasin_raster:Raster):
        super().__init__(reservoir_vector, subbasin_raster)       
        self.__reservoirs = None

    @staticmethod
    def validate(reservoir_vector:Vector,
                 flow_method:str = None, 
                 flow_data_folder:str = None):

        #make sure the shapefile has required columns
        VectorExtension.check_fields_in_vector(reservoir_vector, ReachBMPReservoir.fields_reservoir)

        if flow_method is None: 
            return

        #check the flow method and the corresponding file
        if flow_method not in ReachBMPReservoir.FLOW_ROUTING_METHODS:
            raise ValueError(f"User-define reservoir flow routing method - {flow_method} is not valid. The valid methods are: {', '.join(ReachBMPReservoir.FLOW_ROUTING_METHODS)}")
        
        if flow_method in ReachBMPReservoir.FLOW_ROUTING_METHODS_REQUIRING_EXTERNAL_DATA:
            if flow_data_folder is None or not os.path.exists(flow_data_folder):
                raise ValueError(f"Reservoir flow routing method {flow_method} requires external files. Couldn't find data folder {flow_data_folder}.")
            
            #all the external files should be in the flow_data_folder with an extension of csv and has the same name as specified in FILE columns
            _,field_name_id = VectorExtension.check_id(reservoir_vector) 
            for i in range(reservoir_vector.num_records):
                id = int(reservoir_vector.get_attribute_value(i, field_name_id).get_value_as_f64())
                file = reservoir_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_flow_data)
                if file.is_null() or len(file.get_as_string()) <= 0:
                    raise ValueError(f"FILE column is empty for reservoir {id}.")

                external_file = os.path.join(flow_data_folder, f"{file.get_as_string()}.csv")
                if not os.path.exists(external_file):
                    raise ValueError(f"Couldn't find {external_file} for reservoir {id}.")
                
                #check column in the csv file
                if flow_method == ReachBMPReservoir.FLOW_ROUTING_NAME_RATING_CURVE:
                    df = pd.read_csv(external_file)
                    for col in ReachBMPReservoir.RATING_CURVE_COLUMNS:
                        if col not in df.columns:
                            raise ValueError(f"Couldn't find column {col} in {external_file}. It should have three columns: {', '.join(ReachBMPReservoir.RATING_CURVE_COLUMNS)}")

    @property
    def reservoirs(self):
        if self.__reservoirs is None:
            self.__reservoirs = []
           
            _,field_name_id = VectorExtension.check_id(self.bmp_vector)            
            for i in range(self.bmp_vector.num_records):
                id = int(self.bmp_vector.get_attribute_value(i, field_name_id).get_value_as_f64())
                operation = self.bmp_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_operation)
                emergency_surface_area = self.bmp_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_emergency_surface_area)
                emergency_volume = self.bmp_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_emergency_volume)
                principal_surface_area = self.bmp_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_principal_surface_area)
                principal_volume = self.bmp_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_principal_volume)
                file = self.bmp_vector.get_attribute_value(i, ReachBMPReservoir.field_name_reservoir_flow_data)

                reservoir = Reservoir(id)
                if not operation.is_null():
                    reservoir.OPERATION = operation.get_as_string()
                if not emergency_surface_area.is_null():
                    reservoir.SA_EM = emergency_surface_area.get_value_as_f64()
                if not emergency_volume.is_null():
                    reservoir.V_EM = emergency_volume.get_value_as_f64()        
                if not principal_surface_area.is_null():
                    reservoir.SA_PR = principal_surface_area.get_value_as_f64()
                if not principal_volume.is_null():
                    reservoir.V_PR = principal_volume.get_value_as_f64() 
                if not file.is_null():
                    reservoir.FILE = file.get_as_string()      

                self.__reservoirs.append(reservoir)

        return self.__reservoirs



from typing import Any
from sqlalchemy import Column, Integer, REAL, TEXT
from .bmp_table import BMPTable
from ...names import Names

class FlowDiversion(BMPTable):
    __tablename__ = Names.bmp_table_name_flow_diversion
    """Flow diversion ID"""
    ID = Column(Integer, primary_key=True)
    """X coordinates"""
    XLL = Column(REAL)
    """Y coordinates"""
    YLL = Column(REAL)
    """Operation date"""
    OPERATION = Column(TEXT)
    """Diversion source type, 0 - outside watershed e.g. deep groundwater, 1 - reach"""
    Source = Column(Integer)
    """Source ID"""
    Source_ID = Column(Integer)
    """Diversion target type, 0 - outside watershed e.g. deep groundwater, 1 - reach"""
    Target = Column(Integer)
    """Target ID"""
    Target_ID = Column(Integer)
    """Diversion rate type, 1 - divert fraction of water in source, 2 - leave minimum volume in source, 3 - divert rate (m3/s)"""
    div_type = Column(Integer)
    """Constant diversion rate (fraction, m3, or m3/s)"""
    Q_REL = Column(REAL)
    """Flow diverting method, i.e. MEA_DIV - diversion using measured rate, CONS_DIV - diversion using constant rate (as Q_REL), MON_CONS_DIV - diversion using constant rate in monthes specified in start_mon and end_mon"""
    METHOD = Column(TEXT)
    """Sediment diverting method"""
    SEDMETHOD = Column(TEXT)
    """Nutrients diverting method"""
    NUTMETHOD = Column(TEXT)
    """Measured data table name in HydroClimate DB (must specify if selecting MEA_DIV method)"""
    FILE = Column(TEXT)
    """Start diversion month (only functional if selecting MON_CONS_DIV method)"""
    start_mon = Column(Integer)
    """End diversion month (only functional if selecting MON_CONS_DIV method)"""
    end_mon = Column(Integer)
    """Diversion channel length (m)"""
    length = Column(REAL)
    """Diversion channel width (m)"""
    width = Column(REAL)
    """Diversion channel main side slope (m/m)"""
    main_side_slope = Column(REAL)
    """Diversion channel floodplain side slope (m/m)"""
    floodplain_side_slope = Column(REAL)
    """Diversion channel main side slope (m/m)"""
    w_ratio = Column(REAL)
    """Diversion channel depth (m)"""
    depth = Column(REAL)
    """Diversion channel average channel slope (m/m)"""
    slope = Column(REAL)
    """Diversion channel manning coefficient"""
    manning = Column(REAL)
    """Diversion channel flow velocity (m/s)"""
    velocity = Column(REAL)
    t0 = Column(REAL)
    d0 = Column(REAL)
    max_elevation = Column(REAL)
    min_elevation = Column(REAL)
    Ave_elevation = Column(REAL)
    erodibility_factor = Column(REAL)
    cover_factor = Column(REAL)
    bc1 = Column(REAL)
    bc2 = Column(REAL)
    bc3 = Column(REAL)
    bc4 = Column(REAL)
    rk1 = Column(REAL)
    rk2 = Column(REAL)
    rk3 = Column(REAL)
    rk4 = Column(REAL)
    rs1 = Column(REAL)
    rs2 = Column(REAL)
    rs3 = Column(REAL)
    rs4 = Column(REAL)
    rs5 = Column(REAL)

    def __init__(self):
        self.XLL = 0
        self.YLL = 0
        self.OPERATION = "1900-01-01"
        self.Source = 1
        self.Source_ID = 0
        self.Target = 1
        self.Target_ID = 0
        self.div_type = 1
        self.Q_REL = 0
        self.METHOD = "CONS_DIV"
        self.SEDMETHOD = ""
        self.NUTMETHOD = ""
        self.FILE = ""
        self.start_mon = 4
        self.end_mon = 10
        self.length = 0
        self.width = 0
        self.main_side_slope = 0
        self.floodplain_side_slope = 0
        self.w_ratio = 0
        self.depth = 0
        self.slope = 0
        self.manning = 0
        self.velocity = 0
        self.t0 = 0
        self.d0 = 0
        self.max_elevation = 0
        self.min_elevation = 0
        self.Ave_elevation = 0
        self.erodibility_factor = 0
        self.cover_factor = 0
        self.bc1 = 0
        self.bc2 = 0
        self.bc3 = 0
        self.bc4 = 0
        self.rk1 = 0
        self.rk2 = 0
        self.rk3 = 0
        self.rk4 = 0
        self.rs1 = 0
        self.rs2 = 0
        self.rs3 = 0
        self.rs4 = 0
        self.rs5 = 0







from whitebox_workflows import WbEnvironment, Vector, AttributeField, FieldDataType, FieldData, Raster, VectorGeometryType
from io import StringIO
import pandas as pd
import geopandas as gpd
from .names import Names
import logging
logger = logging.getLogger(__name__)

class VectorExtension:

    @staticmethod
    def vector_to_raster(vector:Vector, base_raster:Raster)->Raster:
        wbe = WbEnvironment()
        exist, id_field_name = VectorExtension.check_id(vector)
        if not exist:
            raise ValueError(f"Couldn't find id field in {vector.file_name}.")

        if vector.header.shape_type == VectorGeometryType.Polygon or vector.header.shape_type == VectorGeometryType.PolygonM or vector.header.shape_type == VectorGeometryType.PolygonZ:
            return wbe.vector_polygons_to_raster(input = vector, field_name = id_field_name, base_raster = base_raster)
        elif vector.header.shape_type == VectorGeometryType.Point or vector.header.shape_type == VectorGeometryType.PointM or vector.header.shape_type == VectorGeometryType.PointZ:
            return wbe.vector_points_to_raster(input = vector, field_name = id_field_name, base_raster = base_raster)
        else:
            raise ValueError(f"Wrong shape type with {vector.file_name}.")

    @staticmethod
    def add_id_for_raster_value(vector:Vector)->Vector:
        if len([field.name for field in vector.get_attribute_fields() if field.name == Names.field_name_raster_value]) <= 0:
            return vector

        id_field = AttributeField(Names.field_name_id, FieldDataType.Int, 10, 0)
        vector.add_attribute_field(id_field)

        for i in range(vector.num_records):
            value = vector.get_attribute_value(i, Names.field_name_raster_value).get_value_as_f64()
            vector.set_attribute_value(i,Names.field_name_id,FieldData.new_int(int(value)))
        return vector
    
    @staticmethod
    def decompsite_overlay_id(vector:Vector, name1:str, name2:str, raster1_max:int)->Vector:
        field1 = AttributeField(name1, FieldDataType.Int, 10, 0)
        field2 = AttributeField(name2, FieldDataType.Int, 10, 0)
        vector.add_attribute_field(field1)
        vector.add_attribute_field(field2)

        for i in range(vector.num_records):
            id = int(vector.get_attribute_value(i, Names.field_name_id).get_value_as_f64())
            id1 = int(id % raster1_max)
            id2 = int((id - id1) / raster1_max)
            vector.set_attribute_value(i,name1,FieldData.new_int(id1))
            vector.set_attribute_value(i,name2,FieldData.new_int(id2))
        return vector       

    @staticmethod
    def get_unique_ids(vector:Vector)->list:
        if vector is None:
            return []
        
        exist, _, ids = VectorExtension.check_unique_id(vector)
        if exist:
            return ids
        return []
    
    @staticmethod
    def get_unique_field_value(vector:Vector, field_name:str, type:type = int)->dict:
        exist, name = VectorExtension.check_field_in_vector(vector, field_name)

        if not exist:
            raise ValueError(f"Can't find field {field_name} in {vector.file_name}.")
        
        exist_id,field_name_id = VectorExtension.check_id(vector)
        
        field_values = {}
        for i in range(vector.num_records):
            id = int(vector.get_attribute_value(i, field_name_id).get_value_as_f64())
            field_value = vector.get_attribute_value(i, name)
            if not field_value.is_null():
                if type is int:
                    if int(field_value.get_value_as_f64()) > 0:
                        field_values[id] = int(field_value.get_value_as_f64())
                elif type is str:
                    if(len(field_value.get_as_string()) > 0):
                        field_values[id] = field_value.get_as_string()

        return field_values

    @staticmethod
    def check_id(vector:Vector):
        """check if the vector has ID column"""
        if vector is None:
            return False, ""

        field_names = [field.name.lower() for field in vector.get_attribute_fields()]
        if Names.field_name_id not in field_names:
            return False, ""
        
        field_index = field_names.index(Names.field_name_id)
        field_name = vector.get_attribute_fields()[field_index].name
        return True, field_name

    @staticmethod
    def check_unique_id(vector:Vector):
        """check if the the vector has the ID column and values are unique"""        
        exist,field_name = VectorExtension.check_id(vector)
        if not exist:
            return False, "", []
        
        ids = []
        for i in range(vector.num_records):
            field_data = vector.get_attribute_value(i, field_name)
            if field_data.is_null():
                raise ValueError(f"The id in {vector.file_name} can't be null.")
            id = int(field_data.get_value_as_f64())
            if id in ids:
                return False, "", []
            ids.append(id)

        return True, field_name, ids

    @staticmethod
    def compare_vector_projection(vector1:Vector, vector2:Vector):
        """Check if the raster have same size as the standard raster """
        return vector1.projection == vector2.projection
        
    
    @staticmethod
    def check_vectors(vectors:dict)->bool:
        """Compare all vectors in the dictionary and return true when all of them has the same projection."""
        if len(vectors) <= 1:
            return True
        
        is_same = True
        standard_vector = None
        for key, value in vectors.items():
            logger.info(f"Checking ID column in {value.file_name} ...")
            exist,_,_ = VectorExtension.check_unique_id(value)
            if not exist:
                raise ValueError(f"ID column was not found in {value.file_name} or the ids are not unique.")

            if standard_vector is None:
                standard_vector = value
                continue

            if not VectorExtension.compare_vector_projection(standard_vector, value):
                is_same = False
                raise ValueError(f"The projection of {value.file_name} doesn't match {standard_vector.file_name}. Please check shapefile projection.")

        return is_same
    
    @staticmethod
    def merge_vectors(vectors:list)->Vector:
        if vectors is None or len(vectors) == 0:
            raise ValueError("There is non vectors to merge.")
        
        if len(vectors) == 1:
            return vectors[0]
        
        wbe = WbEnvironment()
        out_att_fields = [AttributeField("id", FieldDataType.Int, 6, 0)]
        out_vector = wbe.new_vector(vectors[0].header.shape_type, out_att_fields, proj=vectors[0].projection)

        fid = 1
        for v in vectors:
            for i in range(v.num_records):
                geom = v[i]
                out_vector.add_record(geom) # Add the record to the output Vector
                out_vector.add_attribute_record([FieldData.new_int(fid)], deleted=False)
                fid = fid + 1

        return out_vector

    @staticmethod
    def save_vector(vector:Vector, destination_file:str):
        wbe = WbEnvironment()
        out_vector = wbe.new_vector(vector.header.shape_type, vector.get_attribute_fields(), proj=vector.projection)

        # Now fill it with the input data
        for i in range(vector.num_records):
            geom = vector[i]
            out_vector.add_record(geom) # Add the record to the output Vector
            out_vector.add_attribute_record(vector.get_attribute_record(i), deleted=False)
        # Finally, save the output file
        wbe.write_vector(out_vector, destination_file)

        #gpd.read_file(vector.file_name).to_file(destination_file)

    @staticmethod
    def check_field_in_vector(vector:Vector, field_name:str):
        """
        check if vector has attribute with given name
        """
        fields = [field.name for field in vector.get_attribute_fields() if field.name.lower() == field_name.lower()]     
        if len(fields) > 0:
            return True, fields[-1]
        else:
            return False, ""
        
    @staticmethod
    def check_fields_in_vector(vector:Vector, field_names:list):
        """check if vector has list of fields"""
        for field in field_names:
            if not VectorExtension.check_field_in_vector(vector, field)[0]:
                raise ValueError(f"Couldn't find column {field} in {vector.file_name}. It should have following columns: {", ".join(list)}.")
        
    @staticmethod
    def vector_polygons_to_raster_with_boarder(polygon_vector:Vector, field_name:str, base_raster:Raster)->Raster:
        nodata = base_raster.configs.nodata
        rows = base_raster.configs.rows
        cols = base_raster.configs.columns

        wbe = WbEnvironment()
        polygon_raster = wbe.vector_polygons_to_raster(input = polygon_vector, field_name=field_name,base_raster=base_raster)
        boarder_raster = wbe.new_raster(base_raster.configs) 
        conflictWetlands = set()
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]

        for index in range(polygon_vector.num_records):
            #let's use wetland id here
            recNum = int(polygon_vector.get_attribute_value(index, field_name).get_value_as_f64())
            points = polygon_vector[index].points
            numPoints = len(points)
            partData = polygon_vector[index].parts
            numParts = len(partData)
            for part in range(numParts):
                startingPointInPart = partData[part]
                if part < numParts - 1:
                    endingPointInPart = partData[part + 1]
                else:
                    endingPointInPart = numPoints
                n = 0
                for i in range(startingPointInPart, endingPointInPart - 1):
                    x1 = base_raster.get_column_from_x(points[i].x)
                    y1 = base_raster.get_row_from_y(points[i].y)

                    x2 = base_raster.get_column_from_x(points[i+1].x)
                    y2 = base_raster.get_row_from_y(points[i+1].y)

                    d = 0

                    dy = abs(y2 - y1)
                    dx = abs(x2 - x1)

                    dy2 = dy << 1
                    dx2 = dx << 1

                    ix = 1 if x1 < x2 else -1
                    iy = 1 if y1 < y2 else -1

                    if dy <= dx:
                        while True:
                            if base_raster[y1, x1] != nodata:
                                if boarder_raster[y1, x1] > 0 and boarder_raster[y1, x1] != recNum:
                                    wetID = int(polygon_raster[y1, x1])
                                    if wetID > 0 and wetID not in conflictWetlands:
                                        conflictWetlands.add(wetID)
                                        for n in range(8):
                                            rowN = y1 + dY[n]
                                            colN = x1 + dX[n]
                                            neighbourID = int(polygon_raster[rowN, colN])
                                            if neighbourID > 0 and neighbourID != wetID:
                                                if neighbourID not in conflictWetlands:
                                                    conflictWetlands.add(neighbourID)
                                else:
                                    n += 1
                                    boarder_raster[y1, x1] = recNum
                            if x1 == x2:
                                break
                            x1 += ix
                            d += dy2
                            if d > dx:
                                y1 += iy
                                d -= dx2
                    else:
                        while True:
                            if base_raster[y1, x1] != nodata:
                                if boarder_raster[y1, x1] > 0 and boarder_raster[y1, x1] != recNum:
                                    wetID = int(polygon_raster[y1, x1])
                                    if wetID > 0 and wetID not in conflictWetlands:
                                        conflictWetlands.add(wetID)
                                        for n in range(8):
                                            rowN = y1 + dY[n]
                                            colN = x1 + dX[n]
                                            neighbourID = int(polygon_raster[rowN, colN])
                                            if neighbourID > 0 and neighbourID != wetID:
                                                if neighbourID not in conflictWetlands:
                                                    conflictWetlands.add(neighbourID)
                                else:
                                    n += 1
                                    boarder_raster[y1, x1] = recNum
                            if y1 == y2:
                                break
                            y1 += iy
                            d += dx2
                            if d > dy:
                                x1 += ix
                                d -= dy2

           
        for row in range(rows):
            for col in range(cols):
                if boarder_raster[row,col] == nodata and polygon_raster[row,col] != nodata:
                    boarder_raster[row,col] = polygon_raster[row,col]

        return boarder_raster





from ...bmp.bmp_type import DefaultScenarioId

class ReachBMPDistribution:
	"""Location of reach bmps. Note that dugout is not part of the reach bmps"""

	def __init__(self, reach):
		self.Scenario = DefaultScenarioId
		self.Reach = reach		
		self.PointSource = 0
		self.FlowDiversion = 0
		self.Reservoir = 0
		self.Wetland = 0
		self.CatchBasin = 0
		self.GrassWaterway = 0
		self.AccessManagement = 0
		self.WaterUse = 0




from sqlalchemy import Column, Integer, Float, String
from .parameter_table import ParameterTable

class SoilLookup(ParameterTable):
    __tablename__ = 'SoilLookup'  # Name of the table
    SOILCODE = Column(Integer, primary_key=True)
    SNAM = Column(String)
    KS0 = Column(Float)
    DEPTH0 = Column(Integer)
    POROSITY0 = Column(Float)
    FC0 = Column(Float)
    P_INDEX0 = Column(Float)
    RM0 = Column(Float)
    WP0 = Column(Float)
    B_DENSITY0 = Column(Float)
    SAND0 = Column(Float)
    CLAY0 = Column(Float)
    SILT0 = Column(Float)
    USLE_K0 = Column(Float)
    INTERFLOW_SCALE_FACTOR0 = Column(Float)
    KS1 = Column(Float)
    DEPTH1 = Column(Float)
    POROSITY1 = Column(Float)
    FC1 = Column(Float)
    P_INDEX1 = Column(Float)
    RM1 = Column(Float)
    WP1 = Column(Float)
    B_DENSITY1 = Column(Float)
    SAND1 = Column(Float)
    CLAY1 = Column(Float)
    SILT1 = Column(Float)
    USLE_K1 = Column(Float)
    INTERFLOW_SCALE_FACTOR1 = Column(Float)
    KS2 = Column(Float)
    POROSITY2 = Column(Float)
    FC2 = Column(Float)
    P_INDEX2 = Column(Float)
    RM2 = Column(Float)
    WP2 = Column(Float)
    B_DENSITY2 = Column(Float)
    SAND2 = Column(Float)
    CLAY2 = Column(Float)
    SILT2 = Column(Float)
    USLE_K2 = Column(Float)
    INTERFLOW_SCALE_FACTOR2 = Column(Float)
    TEXTURE = Column(Float)
    HG = Column(Float)

    @property
    def AverageK(self)->float:
        return (getattr(self, "KS1") + getattr(self, "KS2")) / 2
    
    @property
    def AveragePorosity(self)->float:
        return (getattr(self, "POROSITY1") + getattr(self, "POROSITY2")) / 2    


from .bmp import BMP
from whitebox_workflows import Vector, Raster
from .bmp_type import BMPType, ReachBMPColumnNames
from ..database.bmp.reach_bmp import ReachBMPDistribution 
import pandas as pd

class ReachBMP(BMP):
    def __init__(self, bmp_vector:Vector, subbasin_raster:Raster):
        super().__init__(bmp_vector, subbasin_raster)

    @staticmethod
    def bmp_type_to_name(bmp_type:BMPType)->str:
        if bmp_type not in ReachBMPColumnNames:
            raise ValueError(f"BMP type: {bmp_type} is not a reach bmp.")
        
        return ReachBMPColumnNames[bmp_type]
    
    @staticmethod
    def create_reach_bmp_df(subbasin_ids:list, reach_bmp_type_id_subbasin_dict:dict)->pd.DataFrame:
        """create reach_bmp table"""

        #create empty object
        reach_bmps = {sub: ReachBMPDistribution(sub) for sub in subbasin_ids}
        
        #assign the id assuming the type is same as the column name
        for type, bmp_id_subbasins in reach_bmp_type_id_subbasin_dict.items():
            if len(bmp_id_subbasins) <= 0:
                continue

            #assign the id     
            for id, sub in bmp_id_subbasins.items():
                setattr(reach_bmps[sub], ReachBMP.bmp_type_to_name(type), id)
            
        return pd.DataFrame([vars(rb) for rb in reach_bmps.values()])



from whitebox_workflows import Raster, WbEnvironment
import pandas as pd
from io import StringIO
from .raster_extension import RasterExtension
import logging

logger = logging.getLogger(__name__)

class Lookup():
    """
    lookup table for raster conversion
    """

    def __init__(self, lookup_csv_file:str, original_raster:Raster) -> None:
        self.csv_file = lookup_csv_file
        self.original_raster = original_raster
        
        #load lookup relationship from csv file assuming it has header and the original id and imwebs id is in first and second column
        logger.info(f"Reading {self.csv_file} ...")
        df = pd.read_csv(self.csv_file, usecols=[0, 1], index_col=0)
        self.lookup_dict = df[df.columns[0]].to_dict()
        self.lookup_list = []

        #get unique values in original raster
        logger.info(f"Reading {self.original_raster.file_name} ...")
        wbe = WbEnvironment()
        values = wbe.list_unique_values_raster(original_raster)
        df = pd.read_csv(StringIO(values), sep =",", usecols = [0], dtype = "int")
        self.raster_ids = df[df.columns[0]].to_list()

        #make sure the raster ids is included in the lookup table
        logger.info(f"Checking ids in lookup table ...")
        non_exist_ids = [id for id in self.raster_ids if id not in self.lookup_dict]
        if len(non_exist_ids) > 0:
            raise ValueError(f"Raster contains ids that are not in the lookup table: {non_exist_ids}.")
        
        #could also test that the imwebs ids are available in the soil/landuse parameter table
        
        #create the mapped raster using the lookup list
        logger.info(f"Create reclassified raster ...")
        self.mapped_raster = RasterExtension.reclassify(self.original_raster, self.lookup_dict)

if __name__ == '__main__':
    wbe = WbEnvironment()
    soil_raster = wbe.read_raster(r"C:\Work\imWEBs\test\test_model\watershed\input\soil.tif")
    lookup = Lookup(r"C:\Work\imWEBs\test\test_model\watershed\input\soilLookup.csv", soil_raster)
    wbe.write_raster(lookup.mapped_raster,r"C:\Work\imWEBs\test\test_model\watershed\input\soil_mapped.tif")
                                  
        




from .bmp_management_base import BMPManagementBaseWithYear
from sqlalchemy import Column, Integer, REAL, TEXT
from .bmp_table import BMPTable
from ...names import Names

class TillageManagement(BMPManagementBaseWithYear):
    """Distribution Table for BMP: Tillage management (14)"""

    def __init__(self, TillMon:int,TillDay:int,TillCode:int):
        super().__init__()
        self.TillMon = TillMon
        self.TillDay = TillDay
        self.TillCode = TillCode  


    @staticmethod
    def default_tillage_management():
        default_tillage_for_crops = {}

        default_tillage_for_crops[1] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[2] =[TillageManagement(TillMon = 5,TillDay = 2,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 5,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 21,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)],
        default_tillage_for_crops[3] =[TillageManagement(TillMon = 5,TillDay = 23,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 26,TillCode = 106)
                ,TillageManagement(TillMon = 9,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 9,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[4] =[TillageManagement(TillMon = 6,TillDay = 2,TillCode = 86)
                ,TillageManagement(TillMon = 6,TillDay = 5,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 28,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 18,TillCode = 108)],
        default_tillage_for_crops[5] =[TillageManagement(TillMon = 5,TillDay = 16,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 19,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 21,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)],
        default_tillage_for_crops[6] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 24,TillCode = 108)],
        default_tillage_for_crops[7] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 9,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[8] =[TillageManagement(TillMon = 5,TillDay = 21,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 23,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 5,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[9] =[TillageManagement(TillMon = 4,TillDay = 23,TillCode = 86)
                ,TillageManagement(TillMon = 4,TillDay = 26,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 9,TillDay = 25,TillCode = 16)],
        default_tillage_for_crops[10] =[TillageManagement(TillMon = 5,TillDay = 2,TillCode = 106)
                ,TillageManagement(TillMon = 5,TillDay = 5,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 9,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[11] =[TillageManagement(TillMon = 9,TillDay = 9,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 12,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 13,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 3,TillCode = 108)],
        default_tillage_for_crops[12] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 21,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)],
        default_tillage_for_crops[13] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 8,TillDay = 14,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 4,TillCode = 108)],
        default_tillage_for_crops[14] =[TillageManagement(TillMon = 5,TillDay = 23,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 26,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 28,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 18,TillCode = 108)],
        default_tillage_for_crops[15] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[16] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[17] =[TillageManagement(TillMon = 5,TillDay = 25,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 25,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[18] =[TillageManagement(TillMon = 6,TillDay = 16,TillCode = 86)
                ,TillageManagement(TillMon = 6,TillDay = 19,TillCode = 106)
                ,TillageManagement(TillMon = 8,TillDay = 14,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 4,TillCode = 108)],
        default_tillage_for_crops[19] =[TillageManagement(TillMon = 6,TillDay = 9,TillCode = 108)
                ,TillageManagement(TillMon = 6,TillDay = 12,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[20] =[TillageManagement(TillMon = 5,TillDay = 23,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 26,TillCode = 106)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[21] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[22] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[23] =[TillageManagement(TillMon = 5,TillDay = 14,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 16,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 15,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 3,TillCode = 108)],
        default_tillage_for_crops[24] =[TillageManagement(TillMon = 5,TillDay = 23,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 26,TillCode = 106)
                ,TillageManagement(TillMon = 9,TillDay = 1,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 8,TillCode = 108)],
        default_tillage_for_crops[25] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 106)
                ,TillageManagement(TillMon = 9,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 19,TillCode = 108)],
        default_tillage_for_crops[26] =[TillageManagement(TillMon = 5,TillDay = 16,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 19,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 28,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 18,TillCode = 108)],
        default_tillage_for_crops[27] =[TillageManagement(TillMon = 4,TillDay = 29,TillCode = 108)
                ,TillageManagement(TillMon = 5,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 11,TillDay = 1,TillCode = 108)],
        default_tillage_for_crops[28] =[TillageManagement(TillMon = 5,TillDay = 16,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 19,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 21,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)],
        default_tillage_for_crops[29] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 106)
                ,TillageManagement(TillMon = 8,TillDay = 28,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 18,TillCode = 108)],
        default_tillage_for_crops[30] =[TillageManagement(TillMon = 6,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 6,TillDay = 5,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 18,TillCode = 108)],
        default_tillage_for_crops[31] =[TillageManagement(TillMon = 5,TillDay = 16,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 19,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 21,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)],
        default_tillage_for_crops[32] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 108)
                ,TillageManagement(TillMon = 10,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[33] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 106)
                ,TillageManagement(TillMon = 10,TillDay = 5,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 25,TillCode = 108)],
        default_tillage_for_crops[34] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 28,TillCode = 89)
                ,TillageManagement(TillMon = 10,TillDay = 18,TillCode = 108)],
        default_tillage_for_crops[35] =[TillageManagement(TillMon = 5,TillDay = 9,TillCode = 86)
                ,TillageManagement(TillMon = 5,TillDay = 12,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 14,TillCode = 106)
                ,TillageManagement(TillMon = 10,TillDay = 4,TillCode = 89)],
        default_tillage_for_crops[36] =[TillageManagement(TillMon = 9,TillDay = 2,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 5,TillCode = 108)
                ,TillageManagement(TillMon = 9,TillDay = 21,TillCode = 106)
                ,TillageManagement(TillMon = 10,TillDay = 11,TillCode = 108)]

class TillageParameter(BMPTable):
    """Parameter Table for BMP: Pasture tillage management (36)"""
    __tablename__ = Names.bmp_table_name_tillage_parameter
    ITNUM = Column(Integer, primary_key=True)
    TILLNM = Column(TEXT)
    DESCRIPTION = Column(TEXT)
    EFFMIX = Column(REAL)
    DEPTIL = Column(REAL)
    CNOP_CN2 = Column(REAL)
    PRC = Column(REAL)
    DSC = Column(REAL)



from .bmp import BMP
from whitebox_workflows import Vector, Raster
import pandas as pd

class StructureBMP(BMP):
    def __init__(self, bmp_vector:Vector, subbasin_raster:Raster):
        super.__init__(bmp_vector, subbasin_raster)






from .bmp_management_base import BMPManagementBaseWithYear

class GrazingManagement(BMPManagementBaseWithYear):
    """Distribution Table for BMP: Grazing management (16)"""

    def __init__(self):
        super().__init__()
        """BMP start month"""
        self.GraMon = 9
        """BMP start day"""
        self.GraDay = 1
        """BMP lasting days"""
        self.Days = 45
        """Animal ID"""
        self.Ani_ID = 1
        """Adult animal percentage"""
        self.Ani_adult = 50
        """Grazing density (1/ha)"""
        self.GR_Density = 15
        """Fraction of day animal stay in field"""
        self.DayFra = 0.3
        """Water source type, 0 - outside watershed e.g. deep groundwater, 1 - reach, 2 - reservoir, 3 - catch basin, 4 - groundwater, 5 - wetland, 6 - dugout"""
        self.Source = 1
        """Water source ID"""
        self.SourceID = 0
        """Dugout IDs in the grazing area"""
        self.Dugout_ID = 0
        """0 – accessible, 1 – managed access, 2 – no access"""
        self.Access = 0
        """0 – no fencing, 1- fencing"""
        self.Fencing = 0
        """Percent of animals drinking in streams. Default is 1"""
        self.StreamAniPerc = 1
        """Percent of day time for drinking, default is 0.02 (0.5/24)"""
        self.Drinking_time = 0.02
        """Average bank erodibility change compared to no-fencing """
        self.BankK_Change = 0.3



import h5py
import os
import shutil
import subprocess
from ..names import Names

import logging
logger = logging.getLogger(__name__)

class ParameterH5:
    def __init__(self, parameter_h5_file) -> None:
        if not os.path.exists(parameter_h5_file):
            raise ValueError(f"{parameter_h5_file} doesn't exist.")
        
        self.parameter_h5_file = parameter_h5_file      
        with h5py.File(parameter_h5_file, 'r') as h5_file:
            for group_name in h5_file.keys():
                if group_name == "asc":
                    group = h5_file[group_name]
                    for attr_name, attr_value in group.attrs.items():
                        if attr_name == "CELL_SIZE":
                            self.cell_num = int(attr_value[0])
                        if attr_name == "DX":
                            self.cell_size = attr_value[0]

    @staticmethod
    def generate_parameter_h5(scenario_folder:str, output_folder:str):
        """
        generate parameter h5 file

        1. Assume all the weight files have been generated in output_folder
        2. run imwebsh5.exe tool to generate parameter.h5 in watershed/output folder
        3. move parameter.h5 to model folder
        4. clean weight files in watershed/output folder
        """

        if not os.path.exists(scenario_folder):
            raise ValueError(f"Couldn't find {scenario_folder}")
        
        if not os.path.exists(output_folder):
            raise ValueError(f"Couldn't find {output_folder}")
        
        #generate the parameter.h5
        imwebs_h5_exe = os.path.join(os.path.dirname(os.path.abspath(__file__)), "engine", "imwebsh5.exe")
        if not os.path.exists(imwebs_h5_exe):
            raise ValueError(f"Couldn't find imwebsh5.exe.")        
        result = subprocess.run([imwebs_h5_exe] + [output_folder], capture_output=True, text=True)
        logger.info(result.stdout)
        logger.info(result.stderr)

        #check if parameter.h5 is generated
        if not os.path.exists(os.path.join(output_folder,Names.parameteH5Name)):
            raise ValueError("parameter.h5 is not generated")

        #copy parameter.h
        shutil.move(os.path.join(output_folder,Names.parameteH5Name),os.path.join(scenario_folder,Names.parameteH5Name))       





import logging
from .config import Config
from configparser import ConfigParser
import logging
from ..model.file_in import FileIn
from ..model.file_out import FileOut
from ..model.config_fig import ConfigFile
import os
from ..database.hydroclimate.hydroclimate_database import HydroClimateDatabase
from ..names import Names
from datetime import date
import pandas as pd
from ..model.parameter_h5 import ParameterH5
from ..interpolation.weight_file import WriteWeightFile
from ..outputs import Outputs
from ..model.model import Model
import numpy as np
import shutil

logger = logging.getLogger(__name__)

class ScenarioConfig(Config):
    """Config for scenario creation"""

    def __init__(self, config_file: str = None):
        super().__init__(config_file)

        if config_file is not None:
            self.__load()

    def __load(self):
        """
        load the config file, validate and copy the input files
        """

        self.data_type_station_ids = {}
        if self.config_file is not None and os.path.exists(self.config_file):           
            logger.info(f"Loading scenario configuration file {self.config_file} ...")
            cf = ConfigParser()
            cf.read(self.config_file)            
            for section, variables in self.config_variables.items():
                for var in variables:
                    value = Config.get_option_value(cf, section, var)  
                    setattr(self, var, None)                  
                    if section == "climate_station" and value is not None and len(value.strip()) > 0:
                        ids = [int(x) for x in value.split(",")]
                        if var.upper() == "T":
                            setattr(self, "TMAX", ids)
                            setattr(self, "TMIN", ids)
                            self.data_type_station_ids["TMAX"] = ids
                            self.data_type_station_ids["TMIN"] = ids
                        else:
                            setattr(self, var, ids)
                            self.data_type_station_ids[var] = ids

                    elif (var == "start_date" or var == "end_date") and value is not None and len(value.strip()) > 0:
                        setattr(self, var, pd.Timestamp(value))
                    elif value is not None and len(value.strip()):
                        setattr(self, var, value)

                    if var == "model_folder":
                        if not os.path.exists(value):
                            raise ValueError(f"Model Folder {value} doesn't exist. Please delineate watershed first.")
                        self.model = Model(value)
                        
                    if var == "name":
                        if value is None or len(value) <=0:
                            raise ValueError("Please give a valide scenario name.")
                        self.scenario_folder = os.path.join(self.model_folder, value)
                        if not os.path.exists(self.scenario_folder):
                            os.makedirs(self.scenario_folder)
            self.__validate()

    def __validate(self):
        #set default value for model type and interval
        if self.model_type is None:
            self.model_type = "cell"
        if self.interval is None:
            self.interval = "daily"

        #if user provide ids, we check if the ids are available.
        #if all is empty, we will use all the available ids
        available_stations_dict = self.model.hydroclimate.data_type_station_ids_dictionary
        if len(self.data_type_station_ids) > 0:            
            for datatype, user_ids in self.data_type_station_ids.items():
                station_ids = available_stations_dict[datatype]
                for id in user_ids:
                    if id not in station_ids:
                        raise ValueError(f"Climate Station {id} for Type {datatype} is not available in hydroclimate database.")
        else:
            self.data_type_station_ids = available_stations_dict

        #check start and end date
        #if the start date and end date is not set, use the start and end date from the database
        database_start_date = self.model.hydroclimate.data_start_date
        database_end_date = self.model.hydroclimate.data_end_date
        if self.start_date is not None and self.start_date < database_start_date:
            raise ValueError(f"The earliest date in hydroclimate database is {database_start_date}, earlier than the user specified start date {self.start_date}")
        if self.end_date is not None and self.end_date > database_end_date:
            raise ValueError(f"The latest date in hydroclimate database is {database_end_date}, earlier than the user specified start date {self.end_date}")                
        if self.start_date is None:
            self.start_date = database_start_date
        if self.end_date is None:
            self.end_date = database_end_date

    @property
    def config_variables(self)->str:
        return {
            #model folder
            "model":["model_folder"],
            
            "scenario":
            [
                "name",        #scenario name will be used as the folder name 
                "model_type",  #cell or subarea based
                "interval",    #daily or hourly
                "start_date",  #empty means use the start date from hydroclimate
                "end_date"     #empty means use the end date from hydroclimate 
            ],

            #climate stations
            #the ids of each station should be given here seperated with comma. 
            #The ids should match the ids in hydroclimate database which will be enforced
            #Exmaple:1,2,3,4
            #Empty means use all the stations for that type. 
            #T is for TMIN and TMAX
            "climate_station":
            [
                "P",
                "T",
                "RM",
                "SR",
                "WS",
                "WD"
            ],

            "climate_interpolation":
            [
                #average_uniform
                #grid_interpolation
                #inverse_distance
                #linear_triangle
                #thiessen_polygon
                "method"
            ]
            }    
    
    def __generate_file_in(self):
        logger.info("Creating file.in ...")
        #get cell size and number of valid cells for cell-based model
        cell_size = -1
        cell_number = -1
        if self.model_type == "cell":
            cell_size = self.model.outputs.inputs.cell_size
            cell_number = self.model.outputs.number_of_valid_cell
        else:
            #read subarea info some where
            pass

        #create file.in
        file_in = FileIn(folder = self.scenario_folder, 
                         model_type=self.model_type, 
                         cell_size=cell_size,
                         cell_num=cell_number,
                         subarea_num=30,
                         subbasin_num=self.model.outputs.number_of_subbasin,
                         start_date=self.start_date,
                         end_date=self.end_date,
                         data_type_station_ids=self.data_type_station_ids, 
                         interval=self.interval)
        
        file_in.write_file()

    def __generate_file_out(self):
        #create file.out
        logger.info("Creating file.out ...")
        file_out = FileOut(self.scenario_folder)
        file_out.write_file()

    def __generate_config_fig(self):
        #create config.fig
        logger.info("Creating config.fig ...")
        config = ConfigFile(self.scenario_folder)
        config.write_file()

    def __generate_weight_file(self):
        logger.info(f"Creating weight file with method {self.method} ...")
        #create weight files use the climate stations selected by users
        coordinates = self.model.hydroclimate.station_coordinates
        previous_ids = []
        previous_weight_file = ""        
        precipitation_weight_file = ""

        #we will create the weight file in the model output folder to facilitate parameter.h5 generation.
        #we will delete any previous weight file first
        weight_file_folder = self.model.model_output_folder
        for wf in [f for f in os.listdir(weight_file_folder) if "weight" in f and ".txt" in f]:
            os.remove(os.path.join(weight_file_folder, wf))

        #start to write
        for datatype, ids in self.data_type_station_ids.items():             
            #we only need one for temperature so skip for TIMIN 
            #same for wind speed and direction    
            if datatype.upper() == "TMIN" or datatype.upper() == "WD":
                continue            
            
            #get weight file name
            weight_file = f"weight_{datatype.lower()}.txt"
            if datatype.upper() == "TMAX":
                weight_file = "weight_t.txt"
            if datatype.upper() == "WS":
                weight_file = "weight_w.txt"

            logger.info(f"Writing {weight_file}")

            #just copy previous weight file if the stations are the same
            if len(previous_ids) > 0 and np.array_equal(np.sort(previous_ids), np.sort(ids)):
                shutil.copyfile(os.path.join(weight_file_folder, previous_weight_file),os.path.join(weight_file_folder, weight_file))
            else:
                #write the weight file
                WriteWeightFile(self.method, 
                                os.path.join(weight_file_folder, weight_file),
                                self.model.outputs.mask_refined_with_subbasin_raster,
                                [coordinates[id] for id in ids])
            #save it
            previous_ids = ids
            previous_weight_file = weight_file
            if datatype.upper() == "P":
                precipitation_weight_file = weight_file

        #copy weight file p for pet
        if "PET" not in self.data_type_station_ids and len(precipitation_weight_file) > 0:
            logger.info("Writing weight_pet.txt")
            shutil.copyfile(os.path.join(weight_file_folder, precipitation_weight_file),os.path.join(weight_file_folder, "weight_pet.txt"))

    def __generate_parameter_h5(self):
        """
        generate parameter h5 file
        """
        ParameterH5.generate_parameter_h5(self.scenario_folder, self.model.model_output_folder)

    def __generate_reach_parameter(self):
        """
        Copy/create reach parameter from watershed/output folder.
        It has same structure as the weight file. The header must start with #        
        """
        logger.info(f"Creating reachParameter.txt ...")
        reach_parameter_df = self.model.outputs.reach_parameter_df.copy()

        #write to file
        reach_parameter_file = os.path.join(self.scenario_folder, Names.reachParameterTxtName)
        reach_parameter_file_temp = os.path.join(self.scenario_folder, "temp.txt")
        with open(reach_parameter_file,'w') as f:
            f.writelines(f"{len(reach_parameter_df)}\n{len(reach_parameter_df.columns)}\n#")

            reach_parameter_df.to_csv(reach_parameter_file_temp, sep = "\t", index=False)
            with open(reach_parameter_file_temp,'r') as f_temp:
                f.write(f_temp.read())

        os.remove(reach_parameter_file_temp)
    

    def __generate_bmp_database(self):
        """just copy the bmp database from database folder"""
        shutil.copy(os.path.join(self.model.model_database_folder, Names.bmpDatabaseName), 
                    os.path.join(self.scenario_folder, Names.bmpDatabaseName))

    def __run_imwebs(self):
        pass

    def generate_model_structure(self):
        """generate model structure"""
        self.__generate_file_in()
        self.__generate_file_out()
        self.__generate_config_fig()
        self.__generate_weight_file()
        self.__generate_parameter_h5()
        self.__generate_reach_parameter()
        self.__generate_bmp_database()









            






import os
from whitebox_workflows import WbEnvironment, Raster, Vector
from .vector_extension import VectorExtension
import pandas as pd

class FolderBase:
    def __init__(self, folder:str) -> None:
        self.folder = folder
        if not os.path.exists(self.folder):
            raise ValueError(f"Folder {folder} doesn't exist.")

        self.wbe = WbEnvironment()
        self.rasters = {}
        self.vectors = {}
        self.dfs = {}

    def get_file_path(self, filename:str)->str:
        return os.path.join(self.folder,filename)

    def save_df(self, df: pd.DataFrame, file_name:str, overwrite = True):
        file_path = self.get_file_path(file_name)

        if os.path.exists(file_path) and not overwrite:
            raise ValueError(f"File: {file_path} already exists and overwrite is set to false.")
        
        df.to_csv(file_path, index=False)
        self.dfs[file_name] = df

    def save_raster(self, raster:Raster, file_name:str, overwrite = True, reload_from_file = True)->Raster:
        file_path = self.get_file_path(file_name)

        if os.path.exists(file_path) and not overwrite:
            raise ValueError(f"File: {file_path} already exists and overwrite is set to false.")
        
        self.wbe.write_raster(raster, file_path)   
        if reload_from_file:
            raster = self.wbe.read_raster(file_path)     
        self.rasters[file_name] = raster
        return raster

    def save_vector(self, vector:Vector, file_name:str, overwrite = True, reload_from_file = False):
        file_path = self.get_file_path(file_name)

        if os.path.exists(file_path) and not overwrite:
            raise ValueError(f"File: {file_path} already exists and overwrite is set to false.")
        
        VectorExtension.save_vector(vector, file_path)  
        if reload_from_file:
            vector = self.wbe.read_vector(file_path)    
        self.vectors[file_name] = vector

    def find_file(self, filename:str)->str:        
        file_path = os.path.join(self.folder,filename)
        if os.path.exists(file_path):
            return file_path
        
        return None
    
    def get_raster(self, raster_name:str)->Raster:
        """get raster with the given name. if the raster doesn't exist, return none."""
        if raster_name in self.rasters:
            return self.rasters[raster_name]
        else:
            file_name = self.find_file(raster_name)
            if file_name is not None:
                self.rasters[raster_name] = self.wbe.read_raster(file_name)
                return self.rasters[raster_name] 
        
        return None
    
    def get_vector(self, shapefile_name:str)->Vector:
        """get vector with the given name. if the vector doesn't exist, return none."""
        if shapefile_name in self.vectors:
            return self.vectors[shapefile_name]
        else:
            file_name = self.find_file(shapefile_name)
            if file_name is not None:
                self.rasters[shapefile_name] = self.wbe.read_vector(file_name)
                return self.rasters[shapefile_name] 
        
        return None      

    def get_df(self, df_name:str)->pd.DataFrame:
        """get data from the given name. if the data frame doesn't exist, return none."""
        if df_name in self.dfs:
            return self.dfs[df_name]
        else:
            file_name = self.find_file(df_name)
            if file_name is not None:
                self.dfs[df_name] = pd.read_csv(file_name)
                return self.dfs[df_name] 
        
        return None      


from sqlalchemy import Column, Integer
from .bmp_table import BMPTable
from ...names import Names

class OffsiteWateringParameter(BMPTable):
    """Parameter Table for BMP: Off site watering (39)"""
    __tablename__ = Names.bmp_table_name_offsite_watering
    """Off-site watering ID"""
    ID = Column(Integer, primary_key = True)
    """Operation starting year (designed for setting up dugout scenarios)"""
    Year = Column(Integer)
    """Subbasin ID"""
    Subbasin = Column(Integer)
    """Water source type,  1 - reach, 2 - reservoir, 3 - catch basin, 4 - groundwater, 5 - wetland, 6 - dugout"""
    Source = Column(Integer)
    """ID for the source"""
    SourceID = Column(Integer)

    def __init__(self, id:int, subbasin:int):
        self.ID = id
        self.Year = 0
        self.Subbasin = subbasin
        self.SourceID = subbasin
        self.Source = 0




from typing import Any
from sqlalchemy import Column, Integer, TEXT, REAL
from .bmp_table import BMPTable
from ...names import Names

class Reservoir(BMPTable):
    """Parameter Table for BMP: Reservoir (3)"""
    __tablename__ = Names.bmp_table_name_reservoir
    ID = Column(Integer, primary_key=True)
    XLL = Column(REAL)
    YLL = Column(REAL)
    OPERATION = Column(TEXT)
    SA_EM = Column(REAL)
    V_EM = Column(REAL)
    SA_PR = Column(REAL)
    V_PR = Column(REAL)
    INI_S = Column(REAL)
    Q_REL = Column(REAL)
    k_res = Column(REAL)
    OFLOWMN01 = Column(REAL)
    OFLOWMN02 = Column(REAL)
    OFLOWMN03 = Column(REAL)
    OFLOWMN04 = Column(REAL)
    OFLOWMN05 = Column(REAL)
    OFLOWMN06 = Column(REAL)
    OFLOWMN07 = Column(REAL)
    OFLOWMN08 = Column(REAL)
    OFLOWMN09 = Column(REAL)
    OFLOWMN10 = Column(REAL)
    OFLOWMN11 = Column(REAL)
    OFLOWMN12 = Column(REAL)
    OFLOWMX01 = Column(REAL)
    OFLOWMX02 = Column(REAL)
    OFLOWMX03 = Column(REAL)
    OFLOWMX04 = Column(REAL)
    OFLOWMX05 = Column(REAL)
    OFLOWMX06 = Column(REAL)
    OFLOWMX07 = Column(REAL)
    OFLOWMX08 = Column(REAL)
    OFLOWMX09 = Column(REAL)
    OFLOWMX10 = Column(REAL)
    OFLOWMX11 = Column(REAL)
    OFLOWMX12 = Column(REAL)
    METHOD = Column(TEXT)
    SEDMETHOD = Column(TEXT)
    INT_SC = Column(REAL)
    NSED = Column(REAL)
    D50 = Column(REAL)
    NUTMETHOD = Column(TEXT)
    FILE = Column(TEXT)
    PSETLR = Column(REAL)
    NSETLR = Column(REAL)
    CHLAR = Column(REAL)
    SECCIR = Column(REAL)
    ORGP = Column(REAL)
    SOLP = Column(REAL)
    ORGN = Column(REAL)
    NO3 = Column(REAL)
    NH3 = Column(REAL)
    NO2 = Column(REAL)
    IFLOD1R = Column(REAL)
    IFLOD2R = Column(REAL)
    NDTARGR = Column(REAL)
    STARG01 = Column(REAL)
    STARG02 = Column(REAL)
    STARG03 = Column(REAL)
    STARG04 = Column(REAL)
    STARG05 = Column(REAL)
    STARG07 = Column(REAL)
    STARG08 = Column(REAL)
    STARG09 = Column(REAL)
    STARG10 = Column(REAL)
    STARG11 = Column(REAL)
    STARG12 = Column(REAL)

    def __init__(self, id:int):
        self.ID = id
        self.XLL = 0
        self.YLL = 0
        self.OPERATION = "1900-01-01"
        self.SA_EM = 0
        self.V_EM = 0
        self.SA_PR = 0
        self.V_PR = 0
        self.INI_S = 0
        self.Q_REL = 0
        self.k_res = 0.3
        self.OFLOWMN01 = 0
        self.OFLOWMN02 = 0
        self.OFLOWMN03 = 0
        self.OFLOWMN04 = 0
        self.OFLOWMN05 = 0
        self.OFLOWMN06 = 0
        self.OFLOWMN07 = 0
        self.OFLOWMN08 = 0
        self.OFLOWMN09 = 0
        self.OFLOWMN10 = 0
        self.OFLOWMN11 = 0
        self.OFLOWMN12 = 0
        self.OFLOWMX01 = 0
        self.OFLOWMX02 = 0
        self.OFLOWMX03 = 0
        self.OFLOWMX04 = 0
        self.OFLOWMX05 = 0
        self.OFLOWMX06 = 0
        self.OFLOWMX07 = 0
        self.OFLOWMX08 = 0
        self.OFLOWMX09 = 0
        self.OFLOWMX10 = 0
        self.OFLOWMX11 = 0
        self.OFLOWMX12 = 0
        self.METHOD = "MMO_RES"
        self.SEDMETHOD = "SMB_RES"
        self.INT_SC = 0.03
        self.NSED = 1
        self.D50 = 10
        self.NUTMETHOD = "RES_NUTR_BAL"
        self.FILE = 0
        self.PSETLR = 50
        self.NSETLR = 10
        self.CHLAR = 1
        self.SECCIR = 1
        self.ORGP = 0
        self.SOLP = 0
        self.ORGN = 0
        self.NO3 = 0
        self.NH3 = 0
        self.NO2 = 0
        self.IFLOD1R = 10
        self.IFLOD2R = 3
        self.NDTARGR = 1
        self.STARG01 = 0.5
        self.STARG02 = 0.5
        self.STARG03 = 0.5
        self.STARG04 = 2
        self.STARG05 = 2
        self.STARG07 = 2
        self.STARG08 = 2
        self.STARG09 = 2
        self.STARG10 = 0.5
        self.STARG11 = 0.5
        self.STARG12 = 0.5
        self.SELECTION = 0



from sqlalchemy import Column, Integer, Float, String, Double
from .bmp_table import BMPTable
from ...names import Names

class Livestock_Parameter(BMPTable):
    __tablename__ = Names.bmp_table_name_livestock_parameter
    """Animal ID"""
    ID = Column(Integer, primary_key=True)
    """Animal name"""
    Name = Column(String)
    """Description"""
    Description = Column(String)
    """Manure ID corresponding the fertilizer table"""
    Man_ID = Column(Integer)
    """Animal weight"""
    Ani_Weight = Column(Double)
    """Fresh manure production per 1000kg live animal mass per day"""
    Man_Day = Column(Double)
    """Dry mass intake per day in percent of body weight"""
    Mass_Intake = Column(Double)
    """Average non-adult weight / adult weight"""
    Ani_nonadult = Column(Double)
    """TSS percentage of the manure """
    Man_TSS_Fra = Column(Double)


from ...delineation.structure_attribute import StructureAttribute

class Wascob:
    """Parameter Table for BMP: WASCob (41)"""
    def __init__(self,attribute:StructureAttribute = None):
        self.Scenario = -1
        self.ID = attribute.id

        self.StartYear = 1900
        self.StartMon = 1
        self.StartDay = 1

        self.FieldId = -1
        self.OutletReachId = -1
        
        self.BermElevation = 0

        self.DeadVolume = 0
        self.DeadArea = 0

        self.NormalVolume = 0
        self.NormalArea = 0

        self.MaxVolume = 0
        self.MaxArea = 0

        self.ContributionArea = attribute.contribution_area
        self.DischargeCapacity = 0

        self.TileOutflowCoefficient = 1
        self.SpillwayDecay = 1

        self.K = 2.5
        self.Nsed = 1
        self.D50 = 10
        self.Dcc = 0.185

        self.PSettle = 10
        self.NSettle = 10
        self.Chlaw = 1
        self.Secciw = 1

        self.InitialVolume = 0
        self.InitialSedimentConc = 0

        self.InitialSolPConc = 0.05
        self.InitialOrgPConc = 0.05

        self.InitialNO3Conc = 0.5
        self.InitialOrgNConc = 0.5
        self.InitialNO2Conc = 0.1
        self.InitialNH3Conc = 0.1

        self.BermType = ''



from typing import Any
from sqlalchemy import Column, Integer, REAL
from .bmp_table import BMPTable
from .bmp_management_base import BMPManagementBaseWithYear
from ...names import Names

class CropManagement(BMPManagementBaseWithYear):    
    """Distribution Table for BMP: Crop management (12)"""
        
    def __init__(self, CropCode:int, PlantingMon:int,PlantingDay:int,HarvestMon:int, HarvestDay:int):
        super().__init__()
        self.CropCode = CropCode
        self.PlantingMon = PlantingMon
        self.PlantingDay = PlantingDay
        self.HarvestMon = HarvestMon
        self.HarvestDay = HarvestDay
        self.HarvestType = 2
        self.HarvestEfficiency = 1
        self.HarvestIndexOverride = 0
        self.StoverFraction = 0
        self.CNOP = 1
        self.IsGrain = 1
        self.PRCOP = 1

    @staticmethod
    def default_crop_management():
        """
        get default crop management for crop 1-36
        """
        default_crop_rotation = {}

        default_crop_rotation[1] = CropManagement(CropCode=1,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[2] = CropManagement(CropCode=2,PlantingMon=5,PlantingDay=4,HarvestMon=8,HarvestDay=18)
        default_crop_rotation[3] = CropManagement(CropCode=3,PlantingMon=5,PlantingDay=25,HarvestMon=9,HarvestDay=14)
        default_crop_rotation[4] = CropManagement(CropCode=4,PlantingMon=6,PlantingDay=4,HarvestMon=10,HarvestDay=11)
        default_crop_rotation[5] = CropManagement(CropCode=5,PlantingMon=5,PlantingDay=18,HarvestMon=9,HarvestDay=11)
        default_crop_rotation[6] = CropManagement(CropCode=6,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[7] = CropManagement(CropCode=7,PlantingMon=5,PlantingDay=1,HarvestMon=8,HarvestDay=18)
        default_crop_rotation[8] = CropManagement(CropCode=8,PlantingMon=5,PlantingDay=21,HarvestMon=10,HarvestDay=25)
        default_crop_rotation[9] = CropManagement(CropCode=9,PlantingMon=4,PlantingDay=25,HarvestMon=10,HarvestDay=18)
        default_crop_rotation[10] = CropManagement(CropCode=10,PlantingMon=5,PlantingDay=4,HarvestMon=10,HarvestDay=4)
        default_crop_rotation[11] = CropManagement(CropCode=11,PlantingMon=9,PlantingDay=4,HarvestMon=8,HarvestDay=11)
        default_crop_rotation[12] = CropManagement(CropCode=12,PlantingMon=5,PlantingDay=11,HarvestMon=10,HarvestDay=11)
        default_crop_rotation[13] = CropManagement(CropCode=13,PlantingMon=5,PlantingDay=1,HarvestMon=8,HarvestDay=4)
        default_crop_rotation[14] = CropManagement(CropCode=14,PlantingMon=5,PlantingDay=25,HarvestMon=9,HarvestDay=18)
        default_crop_rotation[15] = CropManagement(CropCode=15,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[16] = CropManagement(CropCode=16,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[17] = CropManagement(CropCode=17,PlantingMon=5,PlantingDay=25,HarvestMon=8,HarvestDay=11)
        default_crop_rotation[18] = CropManagement(CropCode=18,PlantingMon=6,PlantingDay=18,HarvestMon=9,HarvestDay=4)
        default_crop_rotation[19] = CropManagement(CropCode=19,PlantingMon=6,PlantingDay=11,HarvestMon=8,HarvestDay=25)
        default_crop_rotation[20] = CropManagement(CropCode=20,PlantingMon=5,PlantingDay=25,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[21] = CropManagement(CropCode=21,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[22] = CropManagement(CropCode=22,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[23] = CropManagement(CropCode=23,PlantingMon=5,PlantingDay=14,HarvestMon=9,HarvestDay=6)
        default_crop_rotation[24] = CropManagement(CropCode=24,PlantingMon=5,PlantingDay=25,HarvestMon=7,HarvestDay=25)
        default_crop_rotation[25] = CropManagement(CropCode=25,PlantingMon=5,PlantingDay=11,HarvestMon=9,HarvestDay=25)
        default_crop_rotation[26] = CropManagement(CropCode=26,PlantingMon=5,PlantingDay=18,HarvestMon=9,HarvestDay=18)
        default_crop_rotation[27] = CropManagement(CropCode=27,PlantingMon=5,PlantingDay=1,HarvestMon=10,HarvestDay=1)
        default_crop_rotation[28] = CropManagement(CropCode=28,PlantingMon=5,PlantingDay=18,HarvestMon=9,HarvestDay=25)
        default_crop_rotation[29] = CropManagement(CropCode=29,PlantingMon=5,PlantingDay=11,HarvestMon=8,HarvestDay=18)
        default_crop_rotation[30] = CropManagement(CropCode=30,PlantingMon=6,PlantingDay=4,HarvestMon=9,HarvestDay=14)
        default_crop_rotation[31] = CropManagement(CropCode=31,PlantingMon=5,PlantingDay=18,HarvestMon=9,HarvestDay=11)
        default_crop_rotation[32] = CropManagement(CropCode=32,PlantingMon=5,PlantingDay=11,HarvestMon=10,HarvestDay=11)
        default_crop_rotation[33] = CropManagement(CropCode=33,PlantingMon=5,PlantingDay=11,HarvestMon=9,HarvestDay=25)
        default_crop_rotation[34] = CropManagement(CropCode=34,PlantingMon=5,PlantingDay=11,HarvestMon=9,HarvestDay=11)
        default_crop_rotation[35] = CropManagement(CropCode=35,PlantingMon=5,PlantingDay=11,HarvestMon=8,HarvestDay=18)
        default_crop_rotation[36] = CropManagement(CropCode=36,PlantingMon=9,PlantingDay=4,HarvestMon=8,HarvestDay=11)


        return default_crop_rotation

class CropParameter(BMPTable):
    """Parameter Table for BMP: Pasture crop management (34)"""
    __tablename__ = Names.bmp_table_name_crop_parameter
    ICNUM = Column(Integer, primary_key=True)
    CPNM = Column(REAL)
    DESCRIPTION = Column(REAL)
    IDC = Column(REAL)
    BIO_E = Column(REAL)
    HVSTI = Column(REAL)
    BLAI = Column(REAL)
    FRGRW1 = Column(REAL)
    LAIMX1 = Column(REAL)
    FRGRW2 = Column(REAL)
    LAIMX2 = Column(REAL)
    DLAI = Column(REAL)
    CHTMX = Column(REAL)
    RDMX = Column(REAL)
    T_OPT = Column(REAL)
    T_BASE = Column(REAL)
    CNYLD = Column(REAL)
    CPYLD = Column(REAL)
    BN1 = Column(REAL)
    BN2 = Column(REAL)
    BN3 = Column(REAL)
    BP1 = Column(REAL)
    BP2 = Column(REAL)
    BP3 = Column(REAL)
    WSYF = Column(REAL)
    USLE_C = Column(REAL)
    GSI = Column(REAL)
    VPDFR = Column(REAL)
    FRGMAX = Column(REAL)
    WAVP = Column(REAL)
    CO2HI = Column(REAL)
    BIOEHI = Column(REAL)
    RSDCO_PL = Column(REAL)
    OV_N = Column(REAL)
    CN2A = Column(REAL)
    CN2B = Column(REAL)
    CN2C = Column(REAL)
    CN2D = Column(REAL)
    FERTFIELD = Column(REAL)
    ALAI_MIN = Column(REAL)
    BIO_LEAF = Column(REAL)
    PHU = Column(REAL)
    CNOP = Column(REAL)
    LAI_INIT = Column(REAL)
    BIO_INIT = Column(REAL)
    CURYR_INIT = Column(REAL)

    def __init__(self):
        self.PHU = 1500
        self.CNOP = 1
        self.LAI_INIT = 0.1
        self.BIO_INIT = 10
        self.CURYR_INIT = 10



from sqlalchemy import INT, TEXT, REAL

class RiparianBuffer:
    """Distribution Table for BMP: Riparian buffer (5)"""

    def __init__(self):
        """Scenario ID"""
        self.Scenario = -1
        """Riparian buffer drainage part id used for IMWEBs model calculation, one RIBUF may contain several drainage parts"""
        self.ID = -1
        """Year of operation"""
        self.Year = 1
        """Group riparian buffer ID"""
        self.RIBUF_ID = 0
        """Subbasin ID"""
        self.Subbasin = 0
        """Vegetation ID"""
        self.VegetationID = 6
        """Length of the riparian buffer (m)"""
        self.Length = 0
        """Width of the riparian buffer (m)"""
        self.Width = 0
        """Area of the riparian buffer (ha)"""
        self.Area_ha = 0
        """Drainage area (ha)"""
        self.Drainage_Area = 0
        """Riparian buffer drainage area to riparian buffer area ratio"""
        self.Area_Ratio = 0
        """Slope of the riparian buffer (%)"""
        self.Slope = 0
        """Riparian buffer soil saturation hydraulic conductivity (mm/hr)"""
        self.Sol_K = 0.25
        """Riparian buffer soil porosity"""
        self.Sol_porosity = 0
        """Riparian buffer root depth (mm)"""
        self.Root_Depth = 0

    @staticmethod 
    def column_types()->dict:
        buffer = RiparianBuffer()
        return {col:(INT if col in ["Scenario","ID", "Year","RIBUF_ID","Subbasin","VegetationID"] else REAL) for col in dir(buffer) if "__" not in col}

if __name__ == "__main__":
    col_types = RiparianBuffer.column_types()
    print(len(col_types))



from sqlalchemy import Column, Integer, Float, String
from .parameter_table import ParameterTable

class LanduseLookup(ParameterTable):
    __tablename__ = 'LanduseLookup'
    LANDUSE_ID = Column(Float, primary_key=True)
    CODE = Column(String)
    LANDUSE_NAME = Column(String)
    CN2A = Column(Float)
    CN2B = Column(Float)
    CN2C = Column(Float)
    CN2D = Column(Float)
    ROOT_DEPTH = Column(Float)
    MANNING = Column(Float)
    I_MAX = Column(Float)
    I_MIN = Column(Float)
    SHC = Column(Float)
    SOIL_T10 = Column(Float)
    USLE_C = Column(Float)
    PET_FR = Column(Float)
    PRC_ST1 = Column(Float)
    PRC_ST2 = Column(Float)
    PRC_ST3 = Column(Float)
    PRC_ST4 = Column(Float)
    PRC_ST5 = Column(Float)
    PRC_ST6 = Column(Float)
    PRC_ST7 = Column(Float)
    PRC_ST8 = Column(Float)
    PRC_ST9 = Column(Float)
    PRC_ST10 = Column(Float)
    PRC_ST11 = Column(Float)
    PRC_ST12 = Column(Float)
    SC_ST1 = Column(Float)
    SC_ST2 = Column(Float)
    SC_ST3 = Column(Float)
    SC_ST4 = Column(Float)
    SC_ST5 = Column(Float)
    SC_ST6 = Column(Float)
    SC_ST7 = Column(Float)
    SC_ST8 = Column(Float)
    SC_ST9 = Column(Float)
    SC_ST10 = Column(Float)
    SC_ST11 = Column(Float)
    SC_ST12 = Column(Float)
    DSC_ST1 = Column(Float)
    DSC_ST2 = Column(Float)
    DSC_ST3 = Column(Float)
    DSC_ST4 = Column(Float)
    DSC_ST5 = Column(Float)
    DSC_ST6 = Column(Float)
    DSC_ST7 = Column(Float)
    DSC_ST8 = Column(Float)
    DSC_ST9 = Column(Float)
    DSC_ST10 = Column(Float)
    DSC_ST11 = Column(Float)
    DSC_ST12 = Column(Float)
    FIMP = Column(Float)
    FCIMP = Column(Float)
    CURBDEN = Column(Float)
    URBCOEF = Column(Float)
    DIRTMX = Column(Float)
    THALF = Column(Float)
    TNCONC = Column(Float)
    TPCONC = Column(Float)
    TNO3CONC = Column(Float)
    URBCN2 = Column(Float)
    IS_AGRICULTURAL = Column(Float)
    IsTameGrass = Column(Float)

    @property
    def is_agricultrual(self):
        return getattr(self, "IS_AGRICULTURAL") > 0

    @property
    def field_capacity(self):
        return getattr(self, "FC1")
    
    def getPotentialRunoffCoefficient(self, soil_texture:int)->float:
        """Get potential runoff coefficient for given soil texture"""
        return getattr(self, f"PRC_ST{soil_texture}")

    def getDepressionStorageCapacity(self, soil_texture:int)->float:
        """Get depression storage capacity for given soil texture"""
        return getattr(self, f"DSC_ST{soil_texture}")

    def getSlope(self, soil_texture:int)->float:
        """Get slope for given soil texture"""
        return getattr(self, f"SC_ST{soil_texture}")

    def getCN2(self, hydrological_group)->float:
        """get CN2 for given soil hydrolgical group number 1,2,3,4 which could be read from soil lookup table HG column"""
        soil_classes = ["A", "B", "C", "D"]
        return getattr(self, f"CN2{soil_classes[int(hydrological_group - 1)]}")


from .bmp_management_base import BMPManagementBase

class ManureIncorporationWithin48hManagemet(BMPManagementBase):
    """Distribution Table for BMP: Manure incorporation with 48h (21)"""

    def __init__(self):
        super().__init__()
        self.FerSurface = 0.2




from .bmp_management_base import BMPManagementBase

class ManureSpringApplicationRatherThanFallApplicationManagement(BMPManagementBase):
    """Distribution Table for BMP: Spring application rather than fall application (24)"""
    def __init__(self):
        super().__init__()
